/*
 * RTLinux FIFO interface to GDB stub
 *
 * Written by Michael Barabanov (baraban@fsmlabs.com)
 * Copyright (C) 2000 VJY Associates LLC
 * Released under the terms of the GPL
 *
 */

#include <rtl.h>
#include <rtl_sync.h>
#include <rtl_core.h>
#include <rtl_fifo.h>
#include <rtl_sched.h>
#include <linux/ioctl.h>

extern void set_debug_traps(void);
extern void unset_debug_traps(void);

/* #define RTL_DEBUG_PRINT */
#include <rtl_debug.h>

int fifo = 10;

#define RTL_FIFO_FROM_GDB  (fifo)
#define RTL_FIFO_TO_GDB    (RTL_FIFO_FROM_GDB + 1)
#define RTL_FIFO_FROM_GDB2 (RTL_FIFO_FROM_GDB + 2)

static pthread_t waiting_thread = 0;
#ifdef __SMP__
static spinlock_t fifo_lock = SPIN_LOCK_UNLOCKED;
#endif

int getDebugChar(void)
{
	char c;
	/* blocking FIFOs or mq_receive would be nice here */
	while (1) {
		rtl_irqstate_t flags;
		rtl_spin_lock_irqsave(&fifo_lock, flags);
		if (rtf_get(RTL_FIFO_FROM_GDB2, &c, 1) == 1) {
			rtl_spin_unlock_irqrestore(&fifo_lock, flags);
			break;
		}
/* 		debugpr("S"); */
		pthread_kill (pthread_self(), RTL_SIGNAL_SUSPEND);
		waiting_thread = pthread_self();
		rtl_spin_unlock_irqrestore(&fifo_lock, flags);

		rtl_schedule();
	}

/* 	debugpr("%c", c); */
	return c;
}

void putDebugChar(int chr)
{
	char c = chr;
/* 	debugpr(">%c", c); */
	rtf_put (RTL_FIFO_TO_GDB, &c, 1);
}


pthread_t interrupt_threadid;

static int fifo_handler(unsigned int fifo)
{
	char c;

	if (rtf_get(RTL_FIFO_FROM_GDB, &c, 1) <= 0) {
		return 0;
	}
	if (c == 3)  {
		pthread_wakeup_np(interrupt_threadid);
	} else {
		rtl_irqstate_t flags;
		int cpu = -1;
		do {
			rtf_put(RTL_FIFO_FROM_GDB2, &c, 1);
		} while (rtf_get(RTL_FIFO_FROM_GDB, &c, 1) == 1);

		rtl_spin_lock_irqsave(&fifo_lock, flags);
		if (waiting_thread) {
			cpu = waiting_thread->cpu;
			pthread_kill (waiting_thread, RTL_SIGNAL_WAKEUP);
			waiting_thread = 0;
		}
		rtl_spin_unlock_irqrestore(&fifo_lock, flags);
		if (cpu >= 0) {
#ifdef __SMP__
			if (cpu != rtl_getcpuid())
				rtl_reschedule(cpu);
			else
#endif
				rtl_schedule();
		}
	}
	return 0;
}

static void *interrupt_thread(void *p)
{
	while (1) {
		pthread_suspend_np(pthread_self());
		debugpr("breakpoint\n");
		breakpoint();
	}
	return 0;
}

/* handle flushing */
static int gdb_fifo_ioctl(unsigned int fifo, unsigned int cmd, unsigned long arg)
{
	if (cmd == TCFLSH) {
		if (arg == TCIFLUSH) {
			rtf_flush (fifo + 1);
		}
	}

	return 0;
}


#ifdef MODULE

MODULE_AUTHOR("Michael Barabanov <baraban@fsmlabs.com>");
MODULE_DESCRIPTION("Real-Time Linux Debugger");
MODULE_PARM(fifo, "i");
static int bp = 0;
MODULE_PARM(bp, "i");
static int quiet = 0;
MODULE_PARM(quiet, "i");

int init_module(void)
{
	int ret;
	pthread_attr_t attr;
	struct sched_param sched_param;

	pthread_attr_init (&attr);
	sched_param.sched_priority = sched_get_priority_max(0) + 1;
	pthread_attr_setschedparam (&attr, &sched_param);

	ret = rtf_create(RTL_FIFO_FROM_GDB, 4000);
	ret = rtf_create(RTL_FIFO_TO_GDB, 4000);
	ret = rtf_create(RTL_FIFO_FROM_GDB2, 4000);
	ret = rtf_make_user_pair (RTL_FIFO_FROM_GDB, RTL_FIFO_TO_GDB);
	ret = rtf_link_user_ioctl (RTL_FIFO_FROM_GDB, gdb_fifo_ioctl);
	if (ret < 0) {
		printk("failed to create the RTLinux debugger FIFO (/dev/rtf%d)\n", RTL_FIFO_FROM_GDB);
		printk("ret=%d\n", ret);
		return -1;
	}
	set_debug_traps();
	rtf_create_handler(RTL_FIFO_FROM_GDB, fifo_handler);
	pthread_create (&interrupt_threadid, &attr, interrupt_thread, NULL);
	if ( !quiet ) {
		printk("Real-Time Linux Debugger Loaded (http://www.fsmlabs.com/)\n");
	}
	if (bp) {
		pthread_wakeup_np(interrupt_threadid);
	}

	return 0;
}

void cleanup_module(void)
{
	pthread_delete_np (interrupt_threadid);
	rtf_destroy(RTL_FIFO_FROM_GDB);
	rtf_destroy(RTL_FIFO_TO_GDB);
	rtf_destroy(RTL_FIFO_FROM_GDB2);
	unset_debug_traps();
}

#endif
